! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

!*************************************************************
! Module omsourcemap_mod
!
!
!****************************************************************
module omsourcemap_mod
    use cal
    use dal     
    use OmAllocation_mod     ! For allocating memory for arrays
    use OmSourceParam_mod    ! Defines structures for storing source parameters
    use parameters_mod       ! For computing source moments
    use OmConstants_mod      ! For binx, biny, etc
    use omsmoothimage_mod           ! For polynomial smoothing
    use omnumerics_mod 
    use leastsquares_mod
    use ominterpolate_mod
    use brent_mod
    use cubicshepard_mod
    use splinefit_mod


interface SmoothLevelImage
    module procedure SmoothLevelImage1,  SmoothLevelImage2
end interface SmoothLevelImage


interface DilateImage
    module procedure  DilateLevelImage1, dilateLevelimage2, &
                      DilateImage1, dilateimage2

end interface DilateImage

interface ErodeImage
    module procedure  ErodeImage1, ErodeImage2, ErodeImage3

end interface ErodeImage



interface RefineSourceMap

    module procedure refinesourcemap1, refinesourcemap2, refineSourceMap3

end interface RefineSourceMap

interface RefineExtendedSourceMap

    module procedure refineExtendedsourcemap1, refineExtendedsourcemap2, &
                     refineExtendedsourcemap3

end interface RefineExtendedSourceMap




interface FillInHoles
    module procedure FillInHoles1, FillInHoles2
end interface FillInHoles





!***********************************************
!
!**********************************************
integer(int16), private :: minDx = -2



!logical, private ::  check = .false. ! FOR TESTING PURPOSES
   




contains




subroutine setMinDx( dx )

    integer(int16), intent(in) ::  dx


    mindx = dx
end subroutine setMinDx

!************************************************
!
!***********************************************
subroutine DilateLevelImage1(image, number)
    integer(int16)  :: image(:, :)
    integer(int16), intent(in) :: number
    integer(int16), pointer :: tempImage(:, :) => null()

    integer(int16) :: nx, ny, i, j, il, jl
    integer(int16) :: maxValue

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    il = lbound(image, 1)
    jl = lbound(image, 2)


    tempImage => Allocate2DInt16(il, nx, jl, ny, "tempImage")

    tempImage = image

    do j = jl, ny
        do i = il, nx
            if(image(i, j) .eq. number .or. image(i, j) .ne. 0) cycle
            maxvalue = image(i, j)
            if (i .gt. il) then ! 
                if (image(i - 1, j) .eq. number) maxValue = number ! Adjacent left pixel
             
                if (j .gt. jl) then
                    if (image(i - 1, j - 1) .eq. number) maxvalue = number! Adjacent bottom-left pixel
                end if        
                if (j .lt. ny) then
                    if (image(i - 1, j + 1) .eq. number) maxvalue = number! Adjacent top-left pixel
                end if        

            end if

            if (j .gt. jl) then
            
                if (image(i, j - 1) .eq. number) maxvalue = number! Adjacent bottom pixel
            end if        
            if (j .lt. ny) then
                if (image(i, j + 1) .eq. number) maxvalue = number! Adjacent top pixel
            end if        

           if (i .lt. nx) then ! 
                if (image(i + 1, j) .eq. number) maxvalue = number! Adjacent right pixel
                if (j .gt. jl) then
                    if (image(i + 1, j - 1) .eq. number) maxvalue = number! Adjacent bottom-right pixel
                end if        
                if (j .lt. ny) then
                    if (image(i + 1, j + 1) .eq. number) maxvalue = number! Adjacent top-right pixel
                end if        

            end if
            
            tempImage(i, j) = maxValue
        

        end do
    end do
    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)

end subroutine DilateLevelImage1
!************************************************
!
!***********************************************
subroutine DilateLevelImage2(image, il1, iu1, jl1, ju1, dx, number)
    integer(int16), pointer  :: image(:, :)
    integer(int16), intent(in) :: il1, iu1, jl1, ju1
    integer(int16), intent(in) :: dx, number

    integer(int16), pointer :: tempImage(:, :) => null()

    integer(int16) :: nx, ny, i, j, k, l
    integer(int16) :: maxValue, il, iu, jl, ju

    nx = ubound(image, 1)
    ny = ubound(image, 2)
 
    il = max(0_int16, il1 - dx)
    iu = min(nx, iu1 + dx)
    jl = max(0_int16, jl1 - dx)
    ju = min(ny, ju1 + dx)

    tempImage => Allocate2DInt16(int(il, int32), int(iu, int32), int(jl, int32), int(ju, int32), "tempImage")


    tempImage = image(il : iu, jl : ju)


    do j = jl, ju
        do i = il, iu
            if(image(i, j) .eq. number .or. image(i, j) .ne. 0) cycle
       
            maxvalue = image(i, j)

            kloop : do k = max(0_int16, i - dx), min(nx, i + dx)

                do l = max(0_int16, j - dx), min(ny, j + dx)

                    if (image(k, l) .eq. number) then


                        tempimage(i, j) = number

                        exit kLoop
                    end if
                end do 
            end do kloop

        

        end do
    end do


    image(il : iu, jl : ju) = tempImage


    deallocate(tempImage)
    nullify(tempImage)


end subroutine DilateLevelImage2

!************************************************
!
!
!
!***********************************************
subroutine ErodeImage1(image, number)
    integer(int16)  :: image(:, :)
    integer(int16), intent(in) :: number
 

 
    integer(int16), pointer :: tempImage(:, :) => null()

    integer :: nx, ny, i, j, il, jl
    integer(int16) :: minValue

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    il = lbound(image, 1)
    jl = lbound(image, 2)


    tempImage => Allocate2DInt16(il, nx, jl, ny, "tempImage")

    tempImage = image

    do j = jl, ny
        do i = il, nx
            if (image(i, j) .ne. number) cycle
      
            minValue = image(i, j)
            if (i .gt. il) then ! 
                
                if (j .gt. jl) then
                    if (image(i - 1, j - 1) .eq. 0) minValue = 0! Adjacent bottom-left pixel
                end if        
                if (j .lt. ny) then
                    if (image(i - 1, j + 1) .eq. 0) minValue = 0! Adjacent top-left pixel
                end if        

            end if

            if (j .gt. jl) then
            
                if (image(i, j - 1) .eq. 0) minValue = 0 ! Adjacent bottom pixel
            end if        
            if (j .lt. ny) then
                if (image(i, j + 1) .eq. 0) minValue = 0 ! Adjacent top pixel
            end if        

           if (i .lt. nx) then ! 
                if (image(i + 1, j) .eq. 0) minValue = 0! Adjacent right pixel
                if (j .gt. jl) then
                    if(image(i + 1, j - 1) .eq. 0) minValue = 0 ! Adjacent bottom-right pixel
                end if        
                if (j .lt. ny) then
                    if (image(i + 1, j + 1) .eq. 0) minValue = 0! Adjacent top-right pixel
                end if        

            end if
             
            tempImage(i, j) = minValue
        

        end do
    end do
    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)

end subroutine ErodeImage1


!************************************************
!
!***********************************************
subroutine ErodeImage2(image, il1, iu1, jl1, ju1, number)
    integer(int16)  :: image(:, :)
    
    integer(int16), intent(in) :: il1, iu1, jl1, ju1, number
 

 
    integer(int16), pointer :: tempImage(:, :) => null()

    integer :: nx, ny, i, j, il, jl
    integer(int16) :: minValue

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    il = lbound(image, 1)
    jl = lbound(image, 2)


    tempImage => Allocate2DInt16(il, nx, jl, ny, "tempImage")

    tempImage = image

    do j = jl1, ju1
        do i = il1, iu1
            if (image(i, j) .ne. number) cycle
      
            minValue = image(i, j)
            if (i .gt. il) then ! 
                
                if (j .gt. jl) then
                    if (image(i - 1, j - 1) .eq. 0) minValue = 0! Adjacent bottom-left pixel
                end if        
                if (j .lt. ny) then
                    if (image(i - 1, j + 1) .eq. 0) minValue = 0! Adjacent top-left pixel
                end if        

            end if

            if (j .gt. jl) then
            
                if (image(i, j - 1) .eq. 0) minValue = 0 ! Adjacent bottom pixel
            end if        
            if (j .lt. ny) then
                if (image(i, j + 1) .eq. 0) minValue = 0 ! Adjacent top pixel
            end if        

           if (i .lt. nx) then ! 
                if (image(i + 1, j) .eq. 0) minValue = 0! Adjacent right pixel
                if (j .gt. jl) then
                    if(image(i + 1, j - 1) .eq. 0) minValue = 0! Adjacent bottom-right pixel
                end if        
                if (j .lt. ny) then
                    if (image(i + 1, j + 1) .eq. 0) minValue = 0! Adjacent top-right pixel
                end if        

            end if
             
            tempImage(i, j) = minValue
        

        end do
    end do
    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)

end subroutine ErodeImage2





!***********************************************
!
!***********************************************
subroutine DilateImage1(image, dx)
    real(single), pointer  :: image(:, :)
    integer(int8), intent(in) :: dx
  
    real(single), pointer :: tempImage(:, :) => null()

    integer :: nx, ny, i, j, k, l, il, jl
    real(single) :: maxValue

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    il = lbound(image, 1)
    jl = lbound(image, 2)


    tempImage => Allocate2DSingle(il, nx, jl, ny, "tempImage")

    tempImage = image

    do j = jl, ny
        do i = il, nx
            maxValue = image(i, j)
            do k = max(il, i - dx), min(nx, i + dx)
                do l = max(jl, j - dx), min(ny, j + dx)
                    maxValue = max(maxValue, image(k, l))
                end do
             end do      
        
        
            tempImage(i, j) = maxValue
        end do
    end do
    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)

end subroutine DilateImage1


!***********************************************
!
!***********************************************
subroutine DilateImage2(image, dx)
    integer(int16), pointer  :: image(:, :)
    integer(int8), intent(in) :: dx
 

 
    integer(int16), pointer :: tempImage(:, :) => null()

    integer :: nx, ny, i, j, k, l, il, jl
    integer(int16) :: maxValue

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    il = lbound(image, 1)
    jl = lbound(image, 2)


    tempImage => Allocate2DInt16(il, nx, jl, ny, "tempImage")

    tempImage = image

    do j = jl, ny
        do i = il, nx
            maxValue = image(i, j)
            do k = max(il, i - dx), min(nx, i + dx)
                do l = max(jl, j - dx), min(ny, j + dx)
                    maxValue = max(maxValue, image(k, l))
                end do
             end do      
        
        
            tempImage(i, j) = maxValue
        end do
    end do
    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)

end subroutine DilateImage2

!***********************************************
!
!***********************************************
subroutine ErodeImage3(image, dx)
    real(single), pointer  :: image(:, :)
    integer(int8), intent(in) :: dx
 
 
    real(single), pointer :: tempImage(:, :) => null()

    integer :: nx, ny, i, j, k, l, il, jl
    real(single) :: minValue

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    il = lbound(image, 1)
    jl = lbound(image, 2)


    tempImage => Allocate2DSingle(il, nx, jl, ny, "tempImage")

    tempImage = image

    do j = jl, ny
        do i = il, nx
            minValue = image(i, j)
            do k = max(il, i - dx), min(nx, i + dx)
                do l = max(jl, j - dx), min(ny, j + dx)
                    minValue = min(minValue, image(k, l))
                end do
             end do      
        
        
            tempImage(i, j) = minValue
        end do
    end do
    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)

end subroutine ErodeImage3
!**********************************************************
!
!**********************************************************
subroutine SmoothLevelImage1(image, number, dx, dy)
    
  
    integer(int16), pointer :: image(:, :)
    integer(int16), intent(in) :: number    
    integer, intent(in) :: dx, dy
    integer i, j, k, l, nx, ny
     
    integer(int16), pointer :: tempImage(:, :)
    
  
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    allocate(tempImage(0 : nx, 0 : ny))
    tempImage = image
   
    do j = 0, ny
        
              
        do i = 0, nx
                     
            if (image(i, j) .eq. number .or. image(i, j) .ne. 0) cycle            
           
            kLoop : do k = max(0, i - dx), min(nx, i + dx)

                do l = max(0, j - dy), min(ny, j + dy)
                    if (k .eq. i .and. l .eq. j) cycle
                    if (tempImage(k, l) .eq. number) then
                        image(i, j) = number
                        exit kLoop
                    end if
                end do

            end do  kLoop
                
            
        end do
        
    end do

 
    deallocate(tempImage)

    nullify(tempImage)
     

end subroutine SmoothLevelImage1



!**********************************************************
!
!**********************************************************
subroutine SmoothLevelImage2(image, dx, dy)
    integer(int16), pointer :: image(:, :)
    integer, intent(in) :: dx, dy
    integer i, j, k, l, nx, ny     
    integer(int16), pointer :: tempImage(:, :)
    
  
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    allocate(tempImage(0 : nx, 0 : ny))
   
   
    tempImage = image
    do j = 0, ny
                     
        do i = 0, nx
                     
            if (image(i, j) .eq. int(0, int16)) cycle            
                
            kLoop : do k = max(0, i - dx), min(nx, i + dx)

                do l = max(0, j - dy), min(ny, j + dy)
                
                    if (image(k, l) .eq. int(0, int16)) then
                        tempImage(k, l) = image(i, j)
                      
                    end if
                end do

            end do  kLoop
                
            
        end do
        
    end do

    image = tempImage
    deallocate(tempImage)

    nullify(tempImage)
     

end subroutine SmoothLevelImage2

!**********************************************************
!
!**********************************************************
subroutine SmoothLevelImage3(image, dx, dy)
    integer(int16), pointer :: image(:, :)
    integer, intent(in) :: dx, dy
    integer i, j, k, l, nx, ny     
    integer(int16), pointer :: tempImage(:, :)
    
  
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    allocate(tempImage(0 : nx, 0 : ny))
   
   
    tempImage = image
    do j = 0, ny
                     
        do i = 0, nx
                     
            if (image(i, j) .ge. int(0, int16)) cycle            
                
            kLoop : do k = max(0, i - dx), min(nx, i + dx)

                do l = max(0, j - dy), min(ny, j + dy)
                
                    if (image(k, l) .eq. int(0, int16)) then
  
                        tempImage(k, l) = 1
                     
                    end if
                end do

            end do  kLoop
                
            
        end do
        
    end do

    image = tempImage
    deallocate(tempImage)

    nullify(tempImage)
     

end subroutine SmoothLevelImage3





!**********************************************************************
! RefineSourceMap1
!
!
!**********************************************************************
subroutine RefineSourceMap1(nSources, sourceList, image, levelImage)
    integer, intent(inout) :: nSources
  
    type(SourceLinkedList), intent(inout):: sourceList  ! The linked-list source-list

    real(single), pointer, dimension(:, :) :: image
    integer(int16), pointer, dimension(:, :) :: levelImage
 
   
    integer(int16) :: i

 
    

    type(sourceParameters), pointer :: source(:) => null()
 

    if (nSources .eq. 0) return

    source => SourceDataArray(sourceList, nSources)
    if (.not. associated(source)) return




    call RefineSourceMap2(nSources, source, image, levelImage)

    call deleteSourceList( sourceList )
   

    do i = 1, nSources
        call addSourceToList( sourceList, source(i) )   

    end do
    deallocate(source)
    nullify(source)

   
    

end subroutine refinesourcemap1



!**********************************************************************
! RefineSourceMap2
!
!
!**********************************************************************
subroutine RefineSourceMap2(nSources, source, image, levelImage)
    integer, intent(inout) :: nSources
  
    type(sourceParameters), pointer :: source(:) 
 
    real(single), pointer, dimension(:, :) :: image
    integer(int16), pointer, dimension(:, :) :: levelImage
 


    integer(int16) :: i
    integer(int16) :: nx, ny
    integer(int16) :: dx 


   


    if (nSources .eq. 0) return
    if (.not. associated(source)) return

    nx = ubound(image, 1)
    ny = ubound(image, 2)


  
    write(errstr,'(a)') "Refining source map"
    call message(AppMsg, Verbose)
   

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    dx = 100 / binx

 
    

!write(*, *) "Refinesourcemap2 line 731"
    do i = 1, nSources
!        write(*, *) "Refining source ", i


    
         
        call RefineSourceMap3(source(i), image, levelImage)
 
    end do

  
end subroutine RefineSourceMap2
!*********************************************************
!
!
!*********************************************************
subroutine GetDlmindx(maxImage, dl, mindx)
    integer(int32), intent(in) :: maxImage
    integer(int16), intent(out) :: dl, mindx





    select case (maxImage)
        case(50000 : )
            dl = 12 / binx
              
        case(25000 : 49999)
            dl = 10 / binx 

        case (15000 : 24999) 
            dl = 8 / binx  

        case (5000 : 14999)

            dl = 6 / binx

        case (1000 : 4999)
            dl = 4 / binx

        case (100 : 999)
            dl = 2 / binx

        case default
            dl = 2 / binx  

    end select    

    mindx = dl + 2 / binx

end subroutine GetDlmindx

!**********************************************************************
! RefineSourceMap3 - For a single source
!
!
!**********************************************************************
subroutine RefineSourceMap3(source, image, &
             levelImage, maskImage)
  
  
    type(sourceParameters), intent(inout) :: source 
 
    real(single), pointer, dimension(:, :) :: image
    integer(int16), pointer, dimension(:, :) :: levelImage
  
    logical(boolean), pointer, dimension(:, :), optional :: maskImage
   
    integer(int16) :: i, j
    integer(int16) :: nx, ny
    integer(int16) :: i0, j0, il, iu, jl, ju
    integer(int16) :: dl, dx1
   
    real(single), parameter :: nstd = 3.0
    integer(int32) :: maxImage
    real(single) :: dx, dy, radius, minRadius
  

!    if(source%number == 289 ) then
!        check = .true.
!    else
!        check = .false.
!    end if

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    dx1 = 2 / binx 
 
    i0 = int(source%xpos - 0.4999)

    i0 = min(i0, nx)
    i0 = max(0_int16, i0)
    j0 = int(source%ypos - 0.4999)
    j0 = min(j0, ny)
    j0 = max(0_int16, j0)


    maxImage = image(i0, j0)
  

    call GetDlmindx(maxImage, dl, mindx)

  
    dx1 = mindx + 10 / binx
    il = max(source%imin - dx1, int(0, int16))
    iu = min(source%imax + dx1, nx)
    jl = max(source%jmin - dx1, int(0, int16))
    ju = min(source%jmax + dx1, ny)



    where( abs( levelImage( il : iu, jl : ju)) .eq. source%number)
       
        levelImage( il : iu, jl : ju) = 0
        maskImage ( il : iu, jl : ju) = .false.
    end where
      



    !************************************
    ! 
    !***********************************
  

    dx1 = mindx + 4 / binx
    il = max(i0 - dx1, int(0, int16))
    iu = min(i0 + dx1, nx)
    jl = max(j0 - dx1, int(0, int16))
    ju = min(j0 + dx1, ny)


    source%imin = nx
    source%imax = 0
    source%jmin = ny
    source%jmax = 0

    minRadius = real(2 / binx) 


    
    do j = jl , ju
        dy = real(j - j0)
        do i = il, iu
         
            dx = real(i - i0) 
            radius = sqrt(dx * dx + dy * dy)

            if(radius .le. minRadius) then 

                levelImage(i, j) = source%number

                source%imin = min( source%imin, i)
                source%imax = max( source%imax, i)
                source%jmin = min( source%jmin, j)
                source%jmax = max( source%jmax, j)
                if ( present(maskImage) ) maskImage(i, j) = .true.

            else
            
                if(levelImage(i, j) .ne. 0) cycle
     

                if(radius .le. real(dl)) then 

                    levelImage(i, j) = source%number

                    source%imin = min( source%imin, i)
                    source%imax = max( source%imax, i)
                    source%jmin = min( source%jmin, j)
                    source%jmax = max( source%jmax, j)
             
    
                    if ( present(maskImage) ) maskImage(i, j) = .true.
               
                elseif(radius .le. real(mindx) ) then
                   
                    levelImage(i, j) = -source%number

                 
                end if 
            end if 

        end do
    end do


    call fillInHoles(source, levelImage)





!if (source%number .eq. 2) then
!write(*, *) "mindx=", mindx
!write(*, *) "levelImage(524, 211) .eq. -1)"
!write(*, *) "Source number=", source%number, " x=", source%xpos, " y=", source%ypos
!write(*, *) "Line 1018 refinesourcemap3 imin=", source%imin, " imax=", source%imax
!write(*, *) "jmin=", source%jmin, " jmax=", source%jmax
!write(*, *) "Number of level image pixels set to ", source%number, "  = ", count(levelImage .eq. source%number)
!write(*, *) "Number of level image pixels set to ", source%number, "  = ", count(levelImage .eq. source%number)
!call saveimage(levelimage, int(nx + 1, int16), int(ny + 1, int16), "levelimage.fits", "testimage")
!call fatal("stopped")
!end if

end subroutine RefineSourceMap3

!**********************************************************************
! RefineSourceMap5 - For a single source
!
!
!**********************************************************************
subroutine RefineSourceMap5(source, image, backgroundImage, &
             backgroundSigmaImage, &
             levelImage, nsigma, exposure, maskImage)
  
  
    type(sourceParameters), intent(inout) :: source 
 
    real(single), pointer, dimension(:, :) :: image, backgroundImage, backgroundSigmaImage
    integer(int16), pointer, dimension(:, :) :: levelImage
    real(single), intent(in) :: nsigma, exposure
    logical(boolean), pointer, dimension(:, :), optional :: maskImage
   
    integer(int16) :: i0, j0, j
    integer(int16) :: nx, ny
    integer(int16) :: il, iu, jl, ju, dx, dx1 
    integer(int16) :: k, l, m, n, dl
    integer(int16) :: i2(2)
    real(single), parameter :: nstd = 3.0
    integer(int32) :: maxImage, count
    real(single) :: minHeight
  

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    dx = 2 / binx 
    i0 = int(source%xpos - 0.4999)

    i0 = min(i0, nx)
    i0 = max(0_int16, i0)
    j0 = int(source%ypos - 0.4999)
    j0 = min(j0, ny)
    j0 = max(0_int16, j0)

    il = max(i0 - dx, int(0, int16))
    iu = min(i0 + dx, nx)
    jl = max(j0 - dx, int(0, int16))
    ju = min(j0 + dx, ny)

    i2 = maxloc(image(il : iu, jl : ju)) 

   
    i0 = i2(1) + il - 1
    j0 = i2(2) + jl - 1

  
    maxImage = image(i0, j0)
    !write(*,*) "L1007 i0=", i0, " j0=", j0, " maxImage=", maxImage  

    call GetDlmindx(maxImage, dl, mindx)

 
    do j = source%imin, source%imax
        do k = source%jmin, source%jmax
            if (levelImage(j, k) .eq. source%number) levelImage(j, k) = 0
        end do
    end do
                           

    dx = 50 / binx ! was 100


    il = max(i0 - dx, int(0, int16))
    iu = min(i0 + dx, nx)
    jl = max(j0 - dx, int(0, int16))
    ju = min(j0 + dx, ny)
 
    minHeight = image(i0, j0)
    maxImage = int(minheight * 1000.0 / exposure)
 
    minheight =  max(1.0, 1.5 * &
           minval( backgroundimage(il : iu, jl : ju), &
                   mask=backgroundImage(il : iu , jl : ju) >= 0.0))
 
    source%imin = nx
    source%imax = 0
    source%jmin = ny
    source%jmax = 0

          
   
    count = 0 

    dx1 = 2 / binx

    do l = 1, dx

        il = max( int(0, int16), i0 - l)
        iu = min( nx, i0 + l)
        jl = max( int(0, int16), j0 - l)
        ju = min( ny, j0 + l)


        do n = jl, ju


            do m = il, iu

                if (l .le. dx1) then

                
                   
                    levelImage(m, n) = source%number
                    source%imin = min( source%imin, m)
                    source%imax = max( source%imax, m)
                    source%jmin = min( source%jmin, n)
                    source%jmax = max( source%jmax, n)
                   
                    count = count + 1

                    if (present(maskImage)) maskImage(m, n) = .true.    


                    cycle
                                                            
                end if             



                if (levelImage(m, n) .ne. 0_int16) cycle


                if ( setLevelImagePixel(source, image, backgroundimage, &
                     backgroundSigmaImage, levelimage, m, n, nsigma ) ) then

                    count = count + 1

                    source%imin = min( source%imin, m)
                    source%imax = max( source%imax, m)
                    source%jmin = min( source%jmin, n)
                    source%jmax = max( source%jmax, n)
                    levelImage(m, n) = source%number
                    if (present(maskImage)) maskImage(m, n) = .true.    


                end if
  
            end do
            
        end do

    end do

    call fillInHoles(source, levelImage)




!if (source%number .eq. 28) then
!call saveimage(levelimage, int(nx + 1, int16), int(ny + 1, int16), "levelimage.fits", "testimage")
!call fatal("stopped")
!end if




    source%noSourcePixels = count





end subroutine RefineSourceMap5



!**********************************************************************
! RefineSourceMap1
!
!
!**********************************************************************
subroutine RefineExtendedSourceMap1(nSources, source, image, &
                            backgroundImage, levelImage)
    integer(int16), intent(in) :: nSources
  
    type(sourceParameters), pointer :: source(:) 
 
    real(single), pointer, dimension(:, :) :: image, backgroundImage
    integer(int16), pointer, dimension(:, :) :: levelImage
   

    integer(int16) :: i
    integer(int16) :: nx, ny
   



    if (nSources .eq. 0) return
    if (.not. associated(source)) return


!    write(*, *) "Refineextendedsourcemap1"

    nx = ubound(image, 1)
    ny = ubound(image, 2)

  
    write(errstr,'(a)') "Refining extended source map"
    call message(AppMsg, Verbose)
   

    nx = ubound(image, 1)
    ny = ubound(image, 2)

 



 

    do i = 1, nSources

        call RefineExtendedSourceMap2(source(i), image, backgroundImage, levelImage)
        call fillInHoles(source(i), levelImage)

 
    end do


  
end subroutine RefineExtendedSourceMap1


!**********************************************************************
! RefineSourceMap3
!
!
!**********************************************************************
subroutine RefineExtendedSourceMap3(nSources, sourceList, image, &
                            backgroundImage, levelImage)
    integer(int16), intent(inout) :: nSources
  
 
    type(SourceLinkedList), intent(inout):: sourceList  ! The linked-list source-list

    real(single), pointer, dimension(:, :) :: image, backgroundImage
    integer(int16), pointer, dimension(:, :) :: levelImage
    type(sourceParameters), pointer :: source(:) => null()
 

    integer(int16) :: i
    integer(int16) :: nx, ny
   



    if (nSources .eq. 0) return





    source => SourceDataArray(sourceList, nSources)

    if (.not. associated(source)) return

    call deleteSourceList( sourceList )
  




    nx = ubound(image, 1)
    ny = ubound(image, 2)

  
    write(errstr,'(a)') "Refining extended source map"
    call message(AppMsg, Verbose)
   

    nx = ubound(image, 1)
    ny = ubound(image, 2)

 



 

    do i = 1, nSources


        call RefineExtendedSourceMap2(source(i), image, backgroundImage, levelImage)

        call addSourceToList( sourceList, source(i) )   

 
    end do

   if(associated(source)) deallocate(source)
  
end subroutine RefineExtendedSourceMap3





!**********************************************************************
! RefineExtendedSourceMap - For extended sources
!
!
!**********************************************************************
subroutine RefineExtendedSourceMap2(source, image, backgroundImage, levelImage)
  
  
    type(sourceParameters), intent(inout) :: source 
 
    real(single), pointer, dimension(:, :) :: image, backgroundImage
    integer(int16), pointer, dimension(:, :) :: levelImage
  
    
   
    integer(int16) :: i, j, nx, ny
    integer(int16) :: il, iu, jl, ju
    integer(int16) :: imin(2), imax(2)
    integer(int32) :: callNo
    real(single), parameter :: nstd = 3.0
    real(single) :: fmin, fMax, height



 return

    if(source%extensionFlag == 0) return    

    nx = ubound(image, 1)
    ny = ubound(image, 2)

  


    !*************************************************************************
    ! Find the location of the minimum pixel value within the source region
    !*************************************************************************
    imin = minloc(image(source%imin : source%imax, source%jmin : source%jmax), &
                 levelimage(source%imin : source%imax, source%jmin : source%jmax) == source%number)

    imin(1) = imin(1) + source%imin - 1
    imin(2) = imin(2) + source%jmin - 1

!write(*, *) "line 1161 refine extendedsourcemap2 sourcenumber=", source%number, " imin=", source%imin, " max=", source%imax, &
!   " jmin=", source%jmin, " max=", source%jmax
!write(*, *) "x=", source%xpos, " y=", source%ypos, " number=", source%number, " extension=", source%extensionflag
!write(*, *) "imin(1)=", imin(1), " imin(2) = ", imin(2)



!if( count( levelimage(source%imin : source%imax, source%jmin : source%jmax) == source%number) <=0 ) then
!call saveimage(levelimage, int(nx + 1, int16), int(ny + 1, int16), "levelimage.fits", "testimage")
!call fatal("bug at line 1174, saving levelimage.fits count=0")
!end if



!    write(*, *) "line 1088 imin(1)=", imin(1), " imin(2)=", imin(2), " image=", image(imin(1), imin(2)), " background=", &
!        backgroundimage( imin(1), imin(2))



    fmin = image(imin(1), imin(2)) -  backgroundimage(imin(1), imin(2))

!write(*, *) "Line 1095 fmin=", fmin


    !*************************************************************************
    ! Find the location of the minimum pixel value within the source region
    !*************************************************************************

    imax = maxloc(image(source%imin : source%imax, source%jmin : source%jmax), &
                  levelimage(source%imin : source%imax, source%jmin : source%jmax) == source%number)

!write(*, *) "nx=", nx, " ny=", ny
    imax(1) = imax(1) + source%imin - 1
    imax(2) = imax(2) + source%jmin - 1

!write(*, *) "imax(1)=", imax(1), " imax(2) = ", imax(2)


!   write(*, *) "line 1102 imax(1)=", imax(1), " imax(2)=", imax(2), " image=", image(imax(1), imax(2)), " background=", &
!      backgroundimage(imax(1), imax(2))

    fmax = image(imax(1), imax(2)) - backgroundimage(imax(1), imax(2))

    

!    write(*, *) "source ", source%number, " x=", source%xpos, " y=", source%ypos
!    write(*, *) "imin=", source%imin, " imax=", source%imax
!    write(*, *) "jmin=", source%jmin, " jmax=", source%jmax
!    write(*, *) "backgroundsubtracted fmin=", fmin, " fmax=", fmax

    if(fmin < 5.0) return

    height = fmin / 2.0 
    

    il = source%imin
    iu = source%imax

    jl = source%jmin
    ju = source%jmax

     
    do j = jl, ju
        do i = il, iu

!write(*, *) "Line 1126 i=", i, " j=", j
             
            if (levelImage(i, j) .ne. source%number ) cycle          
        
            callNo = 0 
            
            call setLevelImagePixel2(source, image, backgroundimage, &
                                    levelimage, i, j, nx, ny, &
                                    height, callNo)
!write(*, *) "Line 1135 i=", i, " j=", j

            
        end do           
     
    end do
  
end subroutine RefineExtendedSourceMap2

!**************************************************************************************
!
!
!
!*********************************************************
recursive subroutine setLevelImagePixel2( source, image,  &
                                backgroundimage,          &
                                levelimage, i, j, nx, ny, &
                                height, &
                                callNo)
    type(sourceParameters), intent(inout):: source
    real(single), pointer, dimension(:, :) :: image, backgroundImage
    integer(int16), pointer, dimension(:, :) :: levelImage
    integer(int16), intent(in) :: i, j, nx, ny
    real(single), intent(in) :: height
    integer(int32), intent(inout) :: callNo



    if(callNo >= 100) return
    callNo = callNo + 1


    if ( NumberOfNeighbours22(levelimage, i, j, source%number) < 2) return


    if( (image(i, j) - backgroundImage(i, j)) < height) return

    if(levelImage(i, j) == 0) then
        levelImage(i, j) = source%number
       

        source%imin = min(source%imin, i )
        source%imax = max(source%imax, i)
        source%jmin = min(source%jmin, j)
        source%jmax = max(source%jmax, j)
        levelImage(i, j) = source%number
    end if         




    if(i > 0) then





        call setLevelImagePixel2(source, image, backgroundimage, &
                                 levelimage, i - 1_int16, j, nx, ny, height, callNo)



        if(j > 0) then

            call setLevelImagePixel2(source, image, backgroundimage, &
                        levelimage, i - 1_int16, j - 1_int16, nx, ny, height, callNo)


            call setLevelImagePixel2(source, image, backgroundimage, &
                        levelimage, i, j - 1_int16, nx, ny, height, callNo)

        end if 

        if(j < ny) then

            call setLevelImagePixel2(source, image, backgroundimage, &
                        levelimage, i - 1_int16, j + 1_int16, nx, ny, height, callNo)



            call setLevelImagePixel2(source, image, backgroundimage, &
                        levelimage, i, j + 1_int16, nx, ny, height, callNo)

        end if 



    end if    
         
    if(i < nx) then




        call setLevelImagePixel2(source, image, backgroundimage, &
                                 levelimage, i + 1_int16, j, nx, ny, height, callNo)

        if(j > 0) then



            call setLevelImagePixel2(source, image, backgroundimage, &
                                  levelimage, i + 1_int16, j - 1_int16, nx, ny, height, callNo)

           

        end if 

        if(j < ny) then



            call setLevelImagePixel2(source, image, backgroundimage, &
                                  levelimage, i + 1_int16, j + 1_int16, nx, ny, height, callNo)

          
        end if 



    end if    
         
 

end subroutine setLevelImagePixel2




!***************************************************************************
! Returns the minimum pixel value above the background
! to be used in the source-detection
!
!***************************************************************************
real(single) Function HeightFunction20(i, j, nSigma, &
                      backgroundimage, backgroundsigmaimage)
    integer, intent(in) :: i, j
    real(single), intent(in) :: nSigma
    real(single), pointer, dimension(:, :) :: backgroundImage, backgroundSigmaImage


    HeightFunction20 = max(7.0, backgroundImage(i, j) + nSigma * backgroundSigmaimage(i, j)) ! Friday was 7.0


end Function  HeightFunction20



!***************************************************
! Function NumberOfNeighbours2
! 
!**************************************************
function NumberOfNeighbours22(levelimagePtr, i, j, number) result(count)
    integer(int16), pointer :: levelimagePtr(:, :)
    integer(int16), intent(in) :: i, j 
    integer(int32), intent(in) :: number 
  
    integer(INT16) :: nx, ny

  
    integer(int8)  :: count
    integer(int16) :: il, iu, jl, ju, i1, j1
    ! Local variables
 
    nx = ubound(levelimagePtr, 1)
    ny = ubound(levelimagePtr, 2)

  
    il = max(0_int16, i - 1_int16)
    jl = max(0_int16, j - 1_int16)

  
    iu = min(nx, i + 1_int16)
    ju = min(ny, j + 1_int16)


 

    count = 0
    do i1 = il, iu
        do j1 = jl, ju
            if(i1 .eq. i .and. j1 .eq. j) cycle
            if (levelImagePtr(i1, j1) .eq. number) count = count + 1
        end do
    end do 
  


end function NumberOfNeighbours22


!**********************************************************************
! RefineSourceMap4 - For bright sources
!
!
!**********************************************************************
subroutine RefineSourceMap4(source, backgroundImage, &
             maskImage, emaskimage)
  
  
    type(sourceParameters), intent(inout) :: source 
 
    real(single), pointer, dimension(:, :) :: backgroundImage
   
    logical, pointer, dimension(:, :) :: maskImage, emaskimage
   
    integer(int16) :: nx, ny
    integer :: maxPos(2)
    integer(int16) :: il, iu, jl, ju, dx 
    integer(int16) :: i0, j0, l, m, n, dl

    real(single), parameter :: nstd = 3.0
    
    nx = ubound(backgroundimage, 1)
    ny = ubound(backgroundimage, 2)


    dx = 12 / binx

    il = max(source%imin - dx, int(0, int16))
    iu = min(source%imax + dx, nx)
    jl = max(source%jmin - dx, int(0, int16))
    ju = min(source%jmax + dx, ny)

    maxPos = maxloc(backgroundimage(il : iu, jl : ju))

    i0 = maxPos(1) - 1 + il
    j0 = maxPos(2) - 1 + jl
    dl = 12 / binx
    dx = 100 / binx



    il = max(source%imin - dx, int(0, int16))
    iu = min(source%imax + dx, nx)
    jl = max(source%jmin - dx, int(0, int16))
    ju = min(source%jmax + dx, ny)


    do l = 1, dx
        il = max(int(0, int16), i0 - l)
        iu = min(nx, i0 + l)
        jl = max(int(0, int16), j0 - l)
        ju = min(ny, j0 + l)


        do n = jl, ju
            do m = il, iu

                if (emaskImage(m, n)) cycle
                if (l .le. dl) then
               
                    maskImage(m, n) = .true.    
                    emaskImage(m, n) = .true.    
                         
                else if (setMaskPixel(i0, j0, backgroundimage, m, n)) then
                                
                    maskImage(m, n) = .true.    
                    emaskImage(m, n) = .true.                 
                end if  
            end do            
        end do

    end do
  

!   call saveimage(levelimage, nx + 1, ny + 1, "levelimage.fits", "testimage")

!call fatal("stopped line 8461 displaying levelimage.fits")


  
end subroutine RefineSourceMap4

!**************************************************************************************
!
!
!
!*************************************************************************************
logical Function setMaskPixel(i0, j0, backgroundimage, i, j)
    integer(int16), intent(in) :: i0, j0
    real(single), pointer :: backgroundImage(:, :)
    integer(int16), intent(in) :: i, j
    real(single) :: xpos, ypos, radius, maxradius, radius1, theta, dx, dy, r, dr
    real(single) :: cosTheta, sinTheta
    real(single) :: xl, xu, yl, yu
    integer :: k
    real(single) :: f1, f2
    real(single) :: xCentre, yCentre
    real(single), parameter :: df = 0.003
!integer, parameter :: ipix = 6, jpix=-434

    xCentre = real(i0) + 0.5  ! Centre x pixel coord
    yCentre = real(j0) + 0.5  ! Centre y pixel coord

    xpos = real(i) + 0.5 ! x-pixel coord
    ypos = real(j) + 0.5 ! y-pixel coord
  
    dx = xpos - xCentre
    dy = ypos - yCentre


    radius1 = sqrt(dx * dx + dy * dy)     ! Distance of current pixel from source origin

    if (radius1 .lt. real(4 / binx)) then
        setMaskPixel = .true.     
        return
    end if

!if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1598" 

    theta = atan2(dy, dx)
    cosTheta = cos(theta)
    sinTheta = sin(theta)


    dr = 1.0 / real(binx)
    maxRadius = radius1 + real(6 / binx)  ! Upper radius limit


    xl = 0.0
    xu = real(size(backgroundimage, 1))
    yl = 0.0
    yu = real(size(backgroundimage, 2))

 
    ! Interpolate at the pixel position and either side of it
 
    radius = 3.0 * dr
     
    xpos = xCentre + radius * cosTheta
    if (xpos .lt. xl .or. xpos .gt. xu) then
        setMaskPixel = .true.
        return
    end if

!if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1628 i0=", i0, " j0=", j0 

    ypos = yCentre + radius * sinTheta

    if (ypos .lt. yl .or. ypos .gt. yu) then 
        setMaskPixel = .true.
        return
    end if

!       if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1637" 


    f1 = Interpolate(backgroundimage, xPos, yPos)
  
    if (f1 .lt. 0.1) then




        setMaskPixel = .false.
        return
    end if

!if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1647" 


    do k = 4, 100
       
        r = real(k) * dr







        if(r .gt. maxradius) then
            setMaskPixel = .true.
            return
        end if
!        if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1656 r=", r 


        xpos = xCentre + r * cosTheta
        if (xpos .lt. xl .or. xpos .gt. xu) exit
        ypos = yCentre + r * sinTheta
        if (ypos .lt. yl .or. ypos .gt. yu) exit
       
        f2 = Interpolate(backgroundimage, xPos, yPos)

!      if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1666 f1=", f1, " f2=", f2 

        if (f2 .lt. 0.1) then




            setMaskPixel = .false.                      
            return
        
        end if

!if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1673 f1=", f1, " f2=", f2 

        
        if (  abs(f1 - f2) / f1  .lt. df ) then



            setMaskPixel = .false.                     
            return
    
        
        end if
!if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1684 f1=", f1, " f2=", f2 

     
        f1 = f2
    end do

    setMaskPixel = .true.
       
end Function setMaskPixel

!***********************************************************
!
!***********************************************************
real(single) Function Interpolate(imagePtr, x, y)
    real(single), pointer :: imagePtr(:, :)
    real(single), intent(in) :: x, y

    real(single) :: xl, xu, yl, yu
    real(single) :: u, t
    integer j0, k0, j1, k1
    integer nx, ny

    nx = size(imagePtr, 1)
    ny = size(imagePtr, 2)

    xl = 0.0
    yl = 0.0
    xu = real(nx)
    yu = real(ny)

    if (x .gt. xl .and. x .lt. xu .and. & 
        y .gt. yl .and. y .lt. yu) then

        

        j0 = max(0, int(x - 0.5))
        j0 = min(nx - 1, j0)
      
        k0 = max(0, int(y - 0.5))
        k0 = min(ny - 1, k0)

        
        j1 = j0 + 1
        k1 = k0 + 1

        if (j1 .ge. nx) then
            t = 0.0
            j1 = nx - 1
        else        
            t = max(0.0, x - real(j0) - 0.5)
        end if

        if (k1 .ge. ny) then
            k1 = ny - 1
            u = 0.0
        else  
            u = max(0.0, y - real(k0) - 0.5)
        end if

      
        Interpolate = &
              (1.0 - t) * (1.0 - u) * imagePtr(j0, k0) + &
               t * (1.0 - u) * imagePtr(j1, k0) + &
               t * u * imagePtr(j1, k1) + &
               (1.0  - t) * u * imagePtr(j0, k1)

        Interpolate = max(0.0, interpolate)

        

    else
        Interpolate = 0.0
    end if



end function Interpolate






!**************************************************************************************
!
!
!
!*********************************************************
logical Function setLevelImagePixel(source, image,            &
    backgroundimage, backgroundSigmaImage, levelimage, i, j, nSigma)
    type(sourceParameters), intent(in):: source

 
    real(single), pointer :: image(:, :)
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)

    integer(int16), pointer :: levelImage(:, :)
    real(single), intent(in) :: nSigma
 
    integer(int16), intent(in) :: i, j
    real(single) :: xpos, ypos, radius, theta, dx, dy, r, dr
    real(single) :: cosTheta, sinTheta
    integer, parameter :: maxl = 500
    real(single) :: xl, xu, yl, yu
    integer :: k
    real(single) :: fb, fs, f1, f2, f3, f4, maxRadius
    real(single) :: constant, fmin
   
!    setLevelImagePixel = .false.
!   return 






!if (check) then
!    write(*, *) "setlevelimagepixel Line 1167"
!end if








!if (check) then
!    write(*, *) "setlevelimagepixel Line 1183"
!end if



    if ( image(i, j) .lt. HeightFunction20(int(i), int(j), &
            nSigma, backgroundimage, backgroundSigmaImage)) then ! WAS 2.0


        setLevelImagePixel = .false.
        return
    end if





!if (check) then
!    write(*, *) "setlevelimagepixel Line 1200"
!end if



 

    if (NumberOfNeighbours22(levelimage, i, j, source%number) .lt. 2) then ! was 2



        setLevelImagePixel = .false.           
        return
    end if



!if ( check ) then
!    write(*, *) "setlevelimagepixel Line 1215"
!end if




    xpos = real(i + 1)  ! x-pixel coord
    ypos = real(j + 1)  ! y-pixel coord
    dx = xpos - source%xpos 
    dy = ypos - source%ypos

    radius = sqrt(dx * dx + dy * dy)     ! Distance of current pixel from source origin



!if (check) then
!    write(*, *) "setlevelimagepixel Line 1231 radois=", radius
!end if






    if (radius .lt. real(4 / binx)) then
        setLevelImagePixel = .true.     
        return
    end if

    maxRadius = radius + real(6 / binx)  ! Upper radius limit


    theta = atan2(dy, dx)
    cosTheta = cos(theta)
    sinTheta = sin(theta)


    dr = 0.5 / real(binx)
    xl = 0.0
    xu = real(size(image, 1))
    yl = 0.0
    yu = real(size(image, 2))





    ! Interpolate at the pixel position and either side of it
   
    constant = 1.0 ! was 1.0

  
!if (check) then
!    write(*, *) "setlevelimagepixel Line 1265, x=", source%xpos, " y=", source%ypos
!   write(*, *) "Interpolated value at x,y=",  Interpolate(image, source%xPos, source%yPos)
!end if




    fmin = 100000.0 

    do k = 4, 100
        r = real(k) * dr
        if(r .gt. radius + dr) then
            setLevelImagePixel = .true.
            return
        end if





!if (check) then
!    write(*, *) "setlevelimagepixel Line 1284"
!end if

        xpos = source%xpos + (r - dr) * cosTheta
        if (xpos .lt. xl .or. xpos .gt. xu) exit


        ypos = source%ypos + (r - dr) * sinTheta
        if (ypos .lt. yl .or. ypos .gt. yu) exit
       
        f1 = Interpolate(image, xPos, yPos)
        fmin = min(fmin, f1)




!if (check) then
!    write(*, *) "setlevelimagepixel Line 1304, f1=", f1, " fmin=", fmin
!    write(*, *) "source%xpos=", source%xpos, " y=", source%ypos
!    write(*, *) "xpos=", xpos, " ypos=", ypos, " f1=", f1
!end if





        if (f1 .gt. fmin + sqrt(f1 + fmin)) then



            setLevelImagePixel = .false.                      
            return
   
        end if


        fb = Interpolate( backgroundimage, xPos, yPos )


!if (check) then
!    write(*, *) "setlevelimagepixel Line 1329, fb=", fb
!end if








        !******************************
        ! Remove following for test
        !*****************************
        fs = sqrt(fb)
        if (f1 .lt. max(5.0, fb + 2.0 * fs)) then        

             setLevelImagePixel = .false.
             return
        end if
       

!if (check) then
!    write(*, *) "setlevelimagepixel Line 1351"
!end if



        xpos = source%xpos + r * cosTheta
        if (xpos .lt. xl .or. xpos .gt. xu) exit
        ypos = source%ypos + r * sinTheta
        if (ypos .lt. yl .or. ypos .gt. yu) exit
       
        f2 = Interpolate(image, xPos, yPos)
      



!if (check) then
!    write(*, *) "setlevelimagepixel Line 1370"
!end if




        if (f2 .gt. f1 + constant * sqrt(f1 + f2)) then   ! Test on Friday

         
            setLevelImagePixel = .false.                      
            return
        end if


!if (check) then
!    write(*, *) "setlevelimagepixel Line 1385"
!end if





        xpos = source%xpos + (r + dr) * cosTheta
        if (xpos .lt. xl .or. xpos .gt. xu) exit
        ypos = source%ypos + (r + dr) * sinTheta
        if (ypos .lt. yl .or. ypos .gt. yu) exit
      
        f3 = Interpolate(image, xPos, yPos)
      


!if (check) then
!    write(*, *) "setlevelimagepixel Line 1402"
!end if



        if (f3 .gt. f1 + constant * sqrt(f1 + f3)) then


         
            setLevelImagePixel = .false.                      
            return
        end if


!if (check) then
!    write(*, *) "setlevelimagepixel Line 1419"
!end if






        xpos = source%xpos + (r + 2.0 * dr) * cosTheta
        if (xpos .lt. xl .or. xpos .gt. xu) exit
        ypos = source%ypos + (r + 2.0 * dr) * sinTheta
        if (ypos .lt. yl .or. ypos .gt. yu) exit
     

        f4 = Interpolate(image, xPos, yPos)
          
!if (check) then
!   write(*, *) "setlevelimagepixel Line 1438, f1=", f1, " f4=", f4, " constant * sqrt(f1 + f4)=", &   
!            constant * sqrt(f1 + f4), " xpos=", xpos, " ypos=", ypos, " radius=", radius 
!end if



        if (f4 .gt. f1 + constant * sqrt(f1 + f4)) then
         
            setLevelImagePixel = .false.                      
            return
        end if
 

!if (check) then
!    write(*, *) "setlevelimagepixel Line 1454"
!end if




    end do

    !******************************************************************
    ! Reached here so likely that this pixedl is part of the source
    !******************************************************************

    setLevelImagePixel = .true.       


end Function setLevelImagePixel



!***********************************************************
!
!**********************************************************
subroutine fillInHoles1(nSources, list, levelImage)

    integer(int16), intent(inout) :: nSources
    
 
    type(SourceLinkedList), intent(inout) :: list 
    integer(int16), pointer, dimension(:, :) :: levelimage
  

  
     
    type(sourceParameters), pointer :: source(:) => null()
    integer :: i
  
    
    if(nsources .eq. 0) return



    !****************************************************
    ! Load the point sources from the linked list
    !****************************************************    
    source => SourceDataArray( list, nSources)




    do i = 1, nSources



        call fillInHoles(source(i), levelImage)
       


    end do

    deallocate(source)
    nullify(source)
end subroutine fillInHoles1
!****************************************************************
!
!***************************************************************
subroutine fillInHoles2(source, levelImage)

    
    type(sourceParameters), intent(in) :: source
    integer(int16), pointer, dimension(:, :) :: levelimage
 

    integer(int16) :: i, j
  

    do i = source%imin + 1_int16, source%imax - 1_int16
            
        do j = source%jmin + 1_int16, source%jmax - 1_int16
            if(levelImage(i, j) .eq. source%number) cycle
    
     
                      
            if (levelImage(i - 1, j) .eq. source%number .and. &
                levelImage(i + 1, j) .eq. source%number .and. &
                levelImage(i, j - 1) .eq. source%number .and. &
                levelImage(i, j + 1) .eq. source%number) then
                        
                levelImage(i, j) = source%number
                  
            end if                   
           
        end do 
       
    end do 

end subroutine fillInHoles2


end module Omsourcemap_mod


