! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!*************************************************************
!
!*************************************************************
MODULE sourceanalysis_mod

    use cal
    use dal     
    use OmAllocation_mod     ! For allocating memory for arrays
    use OmSkyBackground_mod  ! For computing sky background mode
    use cubicshepard_mod
    use OmSourceParam_mod    ! Defines structures for storing source parameters
    use omparameters_mod     ! For computing source moments
    use OmSmoothImage_mod    ! For smoothing image
    use OmConstants_mod      ! For binx, biny, etc
    use ompsf_mod
    use omdetectdetection_mod
IMPLICIT NONE



!***************************************************************
! For extended-source detection
!**************************************************************
integer(int16), pointer, private, dimension(:, :) :: tempelevelImage => null()
real(single), pointer, private, dimension(:, :) :: tempBackgroundImage => null()


integer(int8), private :: minCounts    

type(SourceLinkedList), private, save :: tempESourceList  ! The linked-list point-source list



integer(int16), pointer, private, dimension(:, :) :: templevelImage => null()
integer(int16), pointer, private, dimension(:, :) :: maxtemplevelImage => null()
integer(int32), private :: minNopixels, maxNoPixels

integer(int16), private :: kmax, maxnsources


CONTAINS



!***********************************************************
! Checks extended sources and tries to split each one
! into 2 or more sources 
!
!***********************************************************
subroutine decomposeExtendedSources(image,           &
                                    plevelImage,     &
                                    elevelImage,     & 
                                    eSourceList,     &
                                    nESources,       &
                                    isValidSource)

    real(single), pointer, dimension(:, :) :: image
    integer(int16), pointer, dimension(:, :) :: plevelimage, elevelImage
    type(SourceLinkedList), intent(inout):: eSourceList  ! The linked-list extended-source list
    integer(int16), intent(inout) :: neSources   
    type(sourceParameters), pointer :: esource(:) => null()      ! The extended sources
    integer :: i  

    interface

        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:)
            logical :: status

        end function

    end interface




    if (nESources .eq. 0) return
    write(errstr,'(a, i0, a)') "Reanalysing extended sources (", nESources, " sources)"
    call message(AppMsg, Verbose)
 
 
    nx = ubound(image, 1)
    ny = ubound(image, 2)
    
    ! Load the extended sources
    esource => SourceDataArray(esourceList, neSources)
    call DeleteSourceList(eSourceList)

  

    tempeLevelImage => allocate2dInt16(0_int16, nx, 0_int16, ny, "tempelevelimage")
    tempLevelImage => allocate2dInt16(0_int16, nx, 0_int16, ny, "templevelimage")
  
    tempBackgroundImage => allocate2dSingle(0_int16, nx, 0_int16, ny, "tempBackgroundimage")
   
    kmax = 50
   


    maxtempLevelImage => allocate2dInt16(0_int16, nx, 0_int16, ny, "maxtemplevelimage")

    nx = nx + 1
    ny = ny + 1

    !******************************************************************
    ! Loop through all the extended sources and see if they can be 
    ! decomposed into two or more sources
    !
    !
    !*****************************************************************
    neSources = 0    
    call settitle( "" )

    do i = 1, size(esource)  ! Loop through all the extended sources 
!write(*, *) i, " number=", esource(i)%number, " x=", esource(i)%xpos, " y=", esource(i)%ypos


!if (abs(esource(i)%xpos - 25.0) .gt. 8.0 .or. abs(esource(i)%ypos - 165.0) .gt. 8.0) then

!if(eSource(i)%number .ne. -3) cycle
        if (esource(i)%noSourcePixels .lt. 20 / binX .or.    & ! was 100
            esource(i)%noSourcePixels .gt. 50000 / binX .or. &
            esource(i)%extensionFlag .eq. 0 .or.             &
            btest(esource(i)%qualityFlag, 2)) then  




            ! Don't try decomposing this source
            neSources = neSources + 1


            where(eLevelImage(esource(i)%imin : esource(i)%imax, esource(i)%jmin : esource(i)%jmax) &
                   .eq. esource(i)%number) &
                 tempeLevelImage(esource(i)%imin : esource(i)%imax, esource(i)%jmin : esource(i)%jmax) = -neSources 

            eSource(i)%number = -neSources ! The new source number         

            call AddSourceToList(esourceList, eSource(i))



         
        else

!write(*, *) "Line 294 trying to deconvolve source ", i, " source no=", esource(i)%number
!write(*, *) "x=", esource(i)%xpos, " y=", esource(i)%ypos
            !************************************************************
            ! Try to decompose this source into 2 or more sources
            !***********************************************************
            call decomposeExtendedSource(esource(i),  &
                     image,                           &
                     plevelimage,                     &
                     elevelImage,                     &
                     esourceList,                     & 
                     neSources,                       &           
                     isValidSource)                                                                                 



!write(*, *) "No of pixels =", -nesources, "=", count(tempelevelimage==-nesources)



        end if



!write(*, *) "Line 169 minval tempelevelimage=", minval(tempelevelimage)


    
    end do

!  write(*, *) "line 325 minval elevelimage =", minval(elevelimage)
! write(*, *) "line 325 minval tempelevelimage =", minval(tempelevelimage)
!write(*, *) "No of extended sources=", nesources
    write(errstr,'(a, i0)') "New number of extended sources= ", neSources
    call message(AppMsg, Verbose)

    deallocate(esource)
    nullify(esource)






! call saveimage(elevelimage, nx, ny, "elevelimage.fits", "testimage")
! call saveimage(tempelevelimage, nx, ny, "tempelevelimage.fits", "testimage")
!call fatal("stopped at line 419 displaying elevelimage.fits and tempelevelimage.fits")



            
    deallocate(elevelImage)

    elevelImage => tempElevelImage




    deallocate(tempBackgroundImage, templevelimage)
    nullify(tempBackgroundImage, templevelimage)
              

            
    deallocate(maxtemplevelImage)
    nullify(maxtemplevelImage)
 
 


end subroutine DecomposeExtendedSources


!***********************************************************
! Checks extended sources and tries to split each one
! into 2 or more sources 
!
!***********************************************************
subroutine decomposeExtendedSource(esource,             &
                                   image,               &
                                   plevelimage,         &
                                   elevelImage,         &
                                   eSourceList,         &
                                   nESources,           &
                                   isValidSource)

    type(sourceParameters), intent(inout) :: esource      ! The extended sources
    real(single), pointer :: image(:, :)
    integer(int16), pointer, dimension(:, :) :: plevelimage, elevelImage
    type(SourceLinkedList), intent(inout):: eSourceList   ! The linked-list extended-source list
    integer(int16), intent(inout) :: neSources   
    integer(int16) :: i, j, jmax
    real(single) :: height, minImage, maxImage, heightRange
    real(double) :: sum
    integer(int16) ::nSources    
    integer(int16) :: sourceNumber
    type(sourceParameters), pointer :: tempesource(:) => null()  
    real(single) :: minimumCounts 
    integer(int16) :: nx, ny
    interface

        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:)
            logical  :: status

        end function

    end interface





    minNoPixels = 8 / binx  ! Was 20 / binx
 



   minCounts = 2

!write(*, *) "Line 270 source no=", esource%Number
!write(*, *) "xpos=", esource%xpos, " y=", esource%ypos
!write(*, *) "imin=", esource%imin, " imax=", esource%imax, &
!" jmin=", esource%jmin, " jmax=", esource%jmax, " nopixels=", esource%noSourcePixels            

    nx = size(image, 1)
    ny = size(image, 2)


    if (esource%noSourcePixels .lt. minNoPixels .or. &
!        esource%noSourcePixels .gt. 10000 / binX .or. &
        esource%extensionFlag .eq. 0) then
 
        ! Don't try decomposing this source
        neSources = neSources + 1
        sourceNumber = -nesources

        where(eLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) &
                   .eq. esource%number) &
             tempeLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) = sourceNumber 

        eSource%number = sourceNumber ! The new source number


!write(*, *) "Line 360 adding source no ", esource%number, " to list"
!write(*, *) "xpos=", esource%xpos, " y=", esource%ypos
!write(*, *) "imin=", esource%imin, " imax=", esource%imax, " jmin=", esource%jmin, " jmax=", esource%jmax           
        call AddSourceToList(esourceList, eSource)
        return ! Finished        
    end if 
        
    maxNoPixels = esource%noSourcePixels - minNoPixels  

    minImage = 1000000.0
    maximage = 0.0
 
    sum = 0.0_double
 
    do j = esource%jmin, esource%jmax
        do i = esource%imin, esource%imax
            if (elevelImage(i, j) .ne. esource%number) cycle
            if (image(i, j) .ge. 0.0) sum = sum + dble(image(i, j))
            if (image(i, j) .lt. minImage) then
                minImage = image(i, j)
            else if (image(i, j) .gt. maxImage) then
                maximage = image(i, j)
            end if

        end do
    end do 

    minimumCounts = sum / 10.0_double ! The minimum number of counts allowed in a split source
    call setMinimumCounts( minimumCounts ) 

    ! Set the background image to the minimum value within the extended region
    tempbackgroundImage(esource%imin : esource%imax, esource%jmin : esource%jmax) = minimage


    heightRange = (maxImage - minImage) * 0.75         
    
    heightRange = log10(heightrange) / real(kmax)
   
 
    where(elevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) &
            .eq. esource%number)
        tempLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) = 0
    elsewhere
        templevelimage(esource%imin : esource%imax, esource%jmin : esource%jmax) = 1
    end where
        


! call saveimage(templevelimage, nx, ny, "levelimage.fits", "testimage")
!call fatal("stopped at line 419 displaying levelimage.fits")



    !*********************************************************************************
    ! Find the maximum number of sources that the extended source can be split into
    !*********************************************************************************
    maxNSources = 0
    jMax = 1
 
    maxtempLevelImage = 0


    do j = 1, kmax


        height = 10.0 ** (real(j) * heightRange) 
        if ((minImage + height) .gt. maxImage) exit

!write(*, *) "Iteration ", j, " height=", height 
        ! Run the detection algorithim on the extended source region
        
        nSources = 0
          
        call DetectExtendedSources(nx, ny, esource%imin, esource%imax,             &
                   esource%jmin, esource%jmax,                                     &
                   image, image, tempbackgroundImage,                              &
                   maxImage, plevelImage, tempLevelImage,                          &   
                   tempeSourceList, 1_int16, nSources, minNoPixels, maxNoPixels,   &      
                   minCounts,  height, heightFunction2, IsValidSource,             &
                   int(-1, int8), .false.)            

!write(*, *) "Line 379 nsources=", nsources

        if (nSources .gt. maxNSources) then
            !************************************************************************************
            ! Check that at least two thirds of the pixels have been assigned to sources
            !***********************************************************************************
            if( count(tempLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) < 0) &
                .ge. int( real(esource%noSourcePixels) * 0.5) ) then  ! was 0.7 
                maxNSources = nSources
                jmax = j

!write(*, *) "Line 390"
                maxtempLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) = &
                     tempLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax)
        
                if (associated(tempESource)) deallocate(tempEsource)
                ! Load the extended sources
                tempEsource => SourceDataArray(TempesourceList, nSources)

            
            end if
        end if
!write(*, *) "Line 401"


        where (tempLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) .ne. 1) &
                   tempLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) = 0

            
!write(*, *) "Line 408"



        call DeleteSourceList(tempeSourceList)
   
    end do 

!write(*, *) "Line 416"

    if (maxNSources .lt. 2) then ! Not split into 2 or more sources 
        neSources = neSources + 1
        sourceNumber = -neSources  ! The next source number
        where(eLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) &
               .eq. esource%number) &
                tempeLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) = sourceNumber
        esource%number = sourceNumber


        call AddSourceToList(esourceList, eSource)
        return
    end if
    
       
    write(errstr,'(a, i0, a, f0.2, a, f0.2, a, i0, a)') "Source ", esource%number,   &
           " (", esource%xpos, ", ", esource%ypos,                                   &
           ") has been split into ", maxNSources, " sources"
    call message(AppMsg, Verbose)
                            
         

!    write(errstr,'(a)') "New sources"
!    call message(AppMsg, Verbose)
      
    ! Add to the extended-source list   
    do j = 1, maxNSources            
        neSources = neSources + 1
           
        sourceNumber = -neSources

!write(*, *) "New source ", j, " number=", sourceNumber, " imin=", tempesource(j)%imin, " imax=", tempesource(j)%imax, &
!      " jmin=", tempesource(j)%jmin, " imax=", tempesource(j)%jmax
     

        where(maxtempLevelImage(tempeSource(j)%imin : tempeSource(j)%imax, tempESource(j)%jmin : tempESource(j)%jmax)&
              .eq. tempesource(j)%number) &
               tempeLevelImage(tempeSource(j)%imin : tempeSource(j)%imax, tempESource(j)%jmin : tempESource(j)%jmax) = sourceNumber 

        tempESource(j)%number = sourceNumber


        call AddSourceToList(esourceList, tempeSource(j))

!       write(*, *) "minval tempelevelimage=", minval(tempelevelimage)

 
    end do

!write(*, *) "Finished decomposing source"

    deallocate(tempESource)
    nullify(tempESource)


end subroutine DecomposeExtendedSource



end module sourceanalysis_mod    








