! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!************************************************
! Test program for omdetect- checks the output
!************************************************
!  VNY 2.11.2001 division operations are made more robust
program checkoutput
use dal
use errorhandling

implicit none

    type(DataSetT)  :: listDataSet
    type(TableT)    :: tableDataSet
    type(ColumnT)   :: listColumn1,listColumn2

    real(kind=single), pointer :: xpos(:),ypos(:)
    real, allocatable :: xSource(:), ySource(:)
    integer, dimension(:), allocatable :: index
  
    integer :: nSource
    integer :: nFound
    real :: chiSqr
    integer :: i, j, k, l, nrows
    real :: dist
    logical, allocatable :: Found(:)
    integer :: thisSource
    real :: mindist
    
    integer noFiles
    integer :: jStart, Locate
    external Locate
    character(len=50) :: fileName
    character(100) :: text

    integer nCount, noRejects
  

    !****************************************************
    ! The progam createtest.f90 will create one or more
    ! test images (and lists of the sources that it has 
    ! created (testlistn.lis". 
    ! Omdetect will then produce, for each image,
    ! a source-list file called "testlistn.fts". 
    ! n is 1,2, etc, for the number of tests created by
    ! createtest.f90.
    ! Currently doing two tests- one for omichain images
    ! and one for omfchain images.
    !************************************************** 
    noFiles = 1

!open(unit=1, file="test.dat")
            
    do l = 1, noFiles
       
     
        nCount = 0        
        write(fileName, '(a, i1, a)') "testlist", l, ".lis"
             
        open(20, file = trim(fileName) , status = "old", form = "formatted")
   
!        write(*, *) "Checking output from testlist ", trim(filename), " produced by omdetect"
!        write(1, *) "Checking output from testlist ", trim(filename), " produced by omdetect"
   
        write(fileName, '(a, i1, a)') "testlist", l, ".fts"
 
        read(20, *) nSource! The number of sources that were created for the image

        listDataSet = dataSet(trim(fileName), read)
    
        tableDataSet = table(listDataSet, "SRCLIST")
        nrows = numberofRows(tableDataSet)
      

        write(*, '(a, i4)') "No of sources in test image=", nSource
        if (nrows .eq. 0) call fatal("No sources detected", trim(text))

        listColumn1 = column(tableDataSet, "XPOS",read)

        xpos => real32Data(listColumn1)
        listColumn2 = column(tableDataSet, "YPOS",read)
        ypos => real32Data(listColumn2)
 

        !********************************************
        ! Sort this data into increasing x order
        !*******************************************

        call SortData(nrows, xpos, ypos)
                
        write(*, *) "No sources created for test=", nsource
        if (nrows .gt. (nSource+20))  then  ! NB was 0.3
            write(text, '(a, i4, a, i4)') "No of sources detected =", nrows, & 
                             " actual no of test sources=", nsource
         
            call fatal("wrongNumberOfSources", text)
        end if
 


        allocate(xSource(0 : nSource - 1))
        allocate(ySource(0 : nSource - 1))
      
      

        allocate(Found(0 : nSource + 20))
        allocate(index(0 : nSource + 20))

        Found = .false.

       !**********************************
       ! Read in the source list
       !**********************************

        do i = 0, nSource - 1
            read(20, *) xSource(i), ySource(i)
        end do

        !***********************************************
        ! Sort the sources into increasing x-order
        !***********************************************
        call IndexSort(xSource, index, nSource)


        nFound = 0

        chiSqr = 0.0
        thissource = 0
        
        jStart = 0
        noRejects = 0
        do i = 0, nSource - 1
            k = index(i)

            mindist = 1.0e30
!            jStart = Locate(nRows, xpos, xSource(k), jStart)
!            jStart = max(0, jStart - 20)
            do j = jStart, nrows - 1
                if (found(j)) cycle
              
                dist = (xSource(k) - xpos(j)) ** 2 + (ySource(k) - ypos(j)) ** 2
                dist = sqrt(dist)
                if (dist .lt. mindist) then
                    mindist = dist
                    thissource = j
                end if
              
            end do

            mindist = sqrt(mindist)

             if (mindist .lt. 1.414) then  ! was 1.414

                nFound = nFound + 1
 

                nCount = nCount + 1
                
 
           else
!                write(1, 200) "Source no ", k, " not found (", xSource(k), ",", ySource(k), ")" 
   
            end if

                
             if (nfound .eq. nsource) exit               
        end do


200 format(a, i6, a, f6.1, a, f6.1, a, f6.1, a, f6.1, a, f6.1, a, f6.1, a)
300 format(a, f6.1, a, f6.1, a, f6.1)       

         
   

        if (nFound .gt. (nSource+20) .or. nfound .lt. (nsource - 30))  then  ! NB was 0.3
            write(text, '(a, i4, a, i4, a, i4)') "No of sources detected =", nrows, " no matched=", &
                             nfound, " actual no present=", nsource
         
            call fatal("wrongNumberOfSources", text)
        end if
         
      
        deallocate(xSource)
        deallocate(ySource)
       
        deallocate(Found)
        deallocate(index)

        call release(listColumn1)
        call release(listColumn2)
     

        call release(tableDataSet)
        call release(listDataSet)
    end do
    close(unit=1)
!    close(20)
end


!********************************************
! Subroutine SortData
!*******************************************
subroutine SortData(nRows, xpos, ypos)
use dal    
    integer, intent(in) :: nRows
    real(kind=single), dimension(0 : nRows - 1) :: xpos, ypos
    
    real(kind=single), dimension(:), allocatable :: tempArray
    

    integer, dimension(:), allocatable :: index
    integer i, size
    size = nRows - 1
    allocate(index(0 : size))



    call IndexSort(xpos, index, nRows)

    allocate(tempArray(0 : size))
   
    

    tempArray = xpos
    forall(i = 0 : size) xpos(i) = tempArray(index(i))

    tempArray = ypos
    forall(i = 0 : size) ypos(i) = tempArray(index(i))
     
 
    deallocate(tempArray)
  

    deallocate(index)

end subroutine SortData

!************************************************************
! Subroutine HeapSort
! Sorts  one-dimensional array(1.. size) int ascending order.
! An index of the sorted order relative to the original
! order is returned.
!******************** 

subroutine IndexSort(array1, index, size)
    use dal
    integer, intent(in) :: size
    real(kind=single), intent(in), dimension(0 : size - 1) :: array1
    integer, intent(out), dimension(0 : size - 1) :: index

    integer :: i, ir, j, l, irra
    real(kind=double) :: rra

    do i = 0, size - 1
        index(i) = i
    end do   

return
    if (size < 2) return


    l = size / 2 + 1
    ir = size   
  
    do 
        if (l > 1) then
            l = l - 1
            rra = array1(index(l - 1))   
            irra = index(l - 1)    
        else
            rra = array1(index(ir - 1))        
            irra = index(ir - 1)
          
            index(ir - 1) = index(0)
            ir = ir - 1
            
            if (ir <= 1) then
             
                index(0) = irra
                exit
            end if
        end if

        i = l
        j = l + l
        do while(j <= ir)
       
            if (j < ir) then
                if(array1(index(j - 1)) < array1(index(j))) j = j + 1
            end if
            if (rra < array1(index(j - 1))) then
               
                index(i - 1) = index(j - 1)
                i = j
                j = j + j
            else
                exit
            end if
            
        end do   
          
        index(i - 1) = irra
    end do
  
end subroutine IndexSort


!********************************************
! Function Locate
! Given an array table(1 .. n), and
! given a value x, returns the index such that 
! x is between table(j) and table(j+1)
! The search starts at the position jStart
!********************************************
integer function Locate(size, table1, searchValue, jStart)
    use dal
    implicit none
    integer, intent(in) :: size
    real(kind=single), intent(in), dimension(0 : size - 1) :: table1
    real(kind=single), intent(in) :: searchValue
    integer, intent(inout) :: jStart
    integer :: k, klo, khi

    !********************************************
    ! Check to see if x is less than the minimum
    ! value in the table
    !********************************************
    if (searchValue .lt. table1(0)) then
        Locate = 0
        jStart = 0
        return
    end if
  
    !***************************************
    ! Check to see if x is greater than the 
    ! maximum value in the table
    !***************************************
    if (searchValue .gt. table1(size - 1)) then
        Locate = size - 1
        return
    end if

    klo = jStart + 1
    khi = size

    do while((khi - klo) .gt. 1)
        k = (khi + klo) / 2
        if (table1(k - 1) .gt. searchValue) then
            khi = k
        else
            klo = k
        end if
    end do

    !*********************************
    ! klo and khi now bracket the 
    ! searchvalue
    !*********************************
    Locate = klo - 1
    jStart = klo - 1

end function Locate 




